#!/usr/bin/perl
#
# Vlatko Primorac, 1998 - 2000
# 
# You may freely distribute this program under
# GNU General Public License (GPL)
#

use strict;
use Tk;
use Tk::Dialog;
use Tk::LabEntry;
use Getopt::Std;   # single character options

my %args;          # command line options

getopts("hvpilr:c:d:a:n:ts:", \%args);
my $verbose = $args{v};            # verbose
my $scoreFile = $args{s};          # score file name
my $privateColormap = $args{p};    # private colormap
my $iconify = $args{i};            # start iconified 

#
# define the size of the board
#
my $boardSizeX = 9;
if( defined $args{c} ) { $boardSizeX = $args{c} };
my $boardSizeY = 9;
if( defined $args{r} ) { $boardSizeY = $args{r} };

#
# define if we should do the animation at all
# define the speed of animation - miliseconds between kugel spin updates
#
my $doDeleteAnimation = 1;
my $doMoveAnimation   = 0;
my $animationSpeed = 0.02;

#
# define number of kugels to add and delete
#
my $addKugel  = 3 ;
if( defined $args{a} ) { $addKugel = $args{a} };
my $delKugel  = 4 ;  # current one doesn't count
if( defined $args{d} ) { $delKugel = $args{d} - 1 };
my $setOf     = 7 ;  # number of different kugels to use
if( defined $args{n} ) { $setOf = $args{n} };
my $allowSkip = 0;
if( $args{t} ) { $allowSkip = 1 };
my $landscape = 0;
if( $args{l} ) { $landscape = 1 };  # landscape layout

if( $verbose ) {
    foreach my $arg (sort keys %args){
        print "Option $arg = $args{$arg}\n";
    }
}

#
# print command line help
#
if( $args{h} ) {
    print "\n";
    print "$0 [-p] [-r#] [-c#] [-a#] [-d#] [-t] [-s scorefile] [-v]\n";
    print "  -a #            add by #\n";
    print "  -c #            number of columns\n";
    print "  -d #            delete by #\n";
    print "  -h              print this message\n";
    print "  -i              start iconified\n";
    print "  -l              landscape layout\n";
    print "  -n #            use set of # kugels\n";
    print "  -p              use private colormap\n";
    print "  -r #            number of rows\n";
    print "  -s scorefile    specify alternate scorefile\n";
    print "  -t              allow skip through\n";
    print "  -v              verbose mode\n";
    print "\n";
    print 'This is version 2.4 of the perl/tk based  game I call "kugel",'."\n";
    print 'you can  call it whatever you  please.  The game  is played by'."\n";
    print 'moving  objects  on  the board  and putting  them  into  rows,'."\n";
    print 'columns, diagonals.  Each time you create  a line of  the same'."\n";
    print 'objects  ("kugels", balls...), the score  will  increment with'."\n";
    print 'the  number  of objects  you  accomplished to line up.  In the'."\n";
    print 'same time,  the lined up "kugels"  will  disappear, so you can'."\n";
    print 'continue the game with more maneuver  space on the board.  Any'."\n";
    print 'time  you  fail to create  a line of "kugels",  new ones  will'."\n";
    print 'appear on the board. Once the board is full, the game is over.'."\n";
    print "\n";
    print "The latest and greatest version of the game is available at:\n";
    print "http://www.sunsite.unc.edu/pub/Linux/games/strategy/\n";
    print "\n";
    print "Have fun. :)\n\nVlatko Primorac, 1998-2001\nvlatko\@vlptel.com\n\n";
    exit 0;
}

#
# check for versions of Tk module and Perl itself, warn user if modules are old
#
my $tkVersion = $Tk::VERSION;

if( $verbose ) { print "Using Tk version: $tkVersion\n"; }
if( $verbose ) { print "Using Perl version: $]\n"; }

if( $] < 5.005 ) { 
    warn "This program has been tested with Perl version 5.00502\n".
         "You might have problems trying to use it with older version\n";
}
if( $tkVersion < 800.012 ) { 
    warn "This program has been tested with Tk module release 800.012\n".
         "You might have problems trying to use it with older version\n";
}

if( $verbose ) { print "Starting $0 ...\n" ; }

#
# create the main window
#
my $mainFrame;
if($privateColormap) {
    if( $verbose ) { print "Using private colormap.\n" ; }
    $mainFrame = MainWindow->new(-title => "Kugel", -colormap => "new");
} 
else {
    if( $verbose ) { print "Using default colormap.\n" ; }
    $mainFrame = MainWindow->new(-title => "Kugel"); 
}

if( $iconify ) {
    $mainFrame->iconify();  # start iconified
}

my $board = $mainFrame->Frame(-relief => 'raised', -borderwidth => 2);

#
# define pictures to use
#

my $imagesDir;

# search for "path/prgname" or "path\prgname.pl" or "path/prgname.pl"...
if( $0 =~ /(.*)[\\\/][|\\\/]*/ ) { $imagesDir = "$1" ; }
else { $imagesDir = "." }

if( ! $scoreFile ) { $scoreFile = "$imagesDir/scorefile"; }  # high score file

$imagesDir = "$imagesDir/images";

if( $verbose ) { print "Images directory: $imagesDir\n" ; }

opendir(DIR, $imagesDir)  || die "could not open $imagesDir : $!" ;
my @dir = readdir(DIR);
closedir(DIR);

@dir = sort @dir;
shift @dir; # get rid of .
shift @dir; # get rid of .

foreach my $dir (@dir) { $dir = $imagesDir."/".$dir; }

my %photo;

foreach my $dir (@dir) {
    opendir(DIR, $dir)  || die "could not open $dir : $!" ;
    my @imageFiles = readdir(DIR);
    closedir(DIR);

    @imageFiles = sort @imageFiles;

    shift @imageFiles;  # get rid of .
    shift @imageFiles;  # get rid of ..

    foreach my $imageFile (@imageFiles) {
        if( $verbose ) { print "Dir: $dir, Image: $imageFile\n" ; }

        #
        # create anonymous array and assign to it's refference  :)
        #
        push @{$photo{$dir}}, $board->Photo(-format => "gif",
                                            # -height => 50,
                                            # -width =>  50,
                                            -file => "$dir/$imageFile");}
}

# my @currentSet = @{$photo{'./images/set2'}} ;

my @currentSet;

if( defined $dir[1] ) {                  # if 2nd set exist, choose it
    @currentSet = @{$photo{ $dir[1] }} ;
} elsif ( defined $dir[0] ) {            # otherwise, take the 1st set
    @currentSet = @{$photo{ $dir[0] }} ;
} else {                                          # or die (no images found)
    die "no images found\n"  ;
}

if( $setOf > $#currentSet ) {
    $setOf = $#currentSet;
    if($verbose) { print "currentSet reduced to $#currentSet\n"; }
}

#
# color scheme
#
my $ApplPalette = 'goldenrod1';
$mainFrame->setPalette($ApplPalette);  # application color scheme

my @ApplColors = qw( Gray Gold Orange IndianRed Green SkyBlue);

my $bigFont     = $mainFrame->fontCreate(-family => 'courier', 
                                         -weight => 'bold', 
                                         -size => '18');

my $scoreFont   = $mainFrame->fontCreate(-family => 'courier', 
                                         -weight => 'bold', 
                                         -size => '24' );

my $helpFont   = $mainFrame->fontCreate(-family => 'courier', 
                                        -weight => 'bold', 
                                        -size => '18' );

my $nextFrame = $mainFrame->Frame();
my @topKugel;

#
# the score
#
my $score=0;

my $filled=0;  #number of kugels on the screen

my @square;
my @row;
my $didSkip=0;       # did user ever use skip?
my $selected=undef;  # previously selected position
my @scores=undef;
my $highestScore=0;
my $champion="";

my $showShortcuts = 1;  # show bottom shotrtcuts by default
my $showInfo      = 1;  # show "semaphoreFrame"
my $showNext      = 1;  # show next set of kugels

#
# Table containing all the positions in the game. It will be used to 
# implement the undo/redo, and the save of the current game position.
#
# @gameHistory - globally accessible table containing snapshots
#                it is populated with references to array, each array
#                containing one snapshot
#
# $gameHistoryPosition - globally accessible pointer (integer array index)
#                        pointing to the current position used by undo/redo
#                        functionality. taking the new snapshot will always
#                        truncate the @gameHistory array to the lenght of
#                        $gameHistoryPosition and add the current snapshot
#                        on top of it - after meking the new move, one can
#                        no longer redo the previous move.
#
my @gameHistory=();         # empty table, populated in take_snapshot
my $gameHistoryPosition=-1; # no current snapshot, changed by undo/redo

&get_scores;

my $semaphoreFrame = $mainFrame->Frame();
#
# info frame
#
my $infoFrame = $semaphoreFrame->Frame(-relief => 'raised', -borderwidth => 2);
#
# i won't display this label, but i'll leave it here,
# it might be used it in some other version
#
my $numEmptyLabel=$infoFrame->Button(
    -text => "Empty: ".(sprintf '%02d',($boardSizeX*$boardSizeY-$filled))." ("
             .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
                              /($boardSizeX*$boardSizeY)))
             ."\%)",
    -command=>[\&help],
    -font => $scoreFont);

my $setOfLabel = $infoFrame->Button(
    -text => "Set Of $setOf ",
    -command=>[\&set_dimensions],
    -font => $scoreFont);
# set the helper function for the Set Of button - right click activated
$setOfLabel->bind('<3>' => [\&help_about] );

my $addLabel = $infoFrame->Button(
    -text => "Add by $addKugel ",
    -command=>[\&set_dimensions],
    -font => $scoreFont);
# set the helper function for the Add By button - right click activated
$addLabel->bind('<3>' => [\&help_about] );

my $delLabel = $infoFrame->Button(
    -text => "Delete by ".($delKugel+1) ,
    -command=>[\&set_dimensions],
    -font => $scoreFont);
# set the helper function for the Delete By button - right click activated
$delLabel->bind('<3>' => [\&help_about] );

#
# percentage of squares filled
#
my $numFilledLabel=$infoFrame->Button(
    -text => (sprintf '%02d',
             int(100*$filled/($boardSizeX*$boardSizeY)))."\% Filled",
    -command=>[\&help],
    -font => $scoreFont);

#
# score frame
#
my $scoreFrame = $semaphoreFrame->Frame(-relief => 'raised', -borderwidth => 2);
my $scoreLabel = $scoreFrame->Button(
    -text => "Score: ".(sprintf '%04d',$score),
    -command => [\&show_hall_of_fame],
    -font => $scoreFont);
# set the helper function for the Score button - right click activated
$scoreLabel->bind('<3>' => [\&help_about] );
my $highScoreLabel = $scoreFrame->Button(
    -text => "$champion: ".(sprintf '%04d',$highestScore),
    -command => [\&show_hall_of_fame],
    -font => $scoreFont);
# set the helper function for the High Score button - right click activated
$highScoreLabel->bind('<3>' => [\&help_about] );

#
# button menu shortcuts, down
#
my $bottomFrame = $mainFrame->Frame();

#
# color buttons, part of bottomFrame
#
my $colors = $bottomFrame->Frame();
foreach my $color (@ApplColors) {
    $colors->Button(
        -text   => $color,
        -pady => '0',
        -padx => '0',
        -background => $color,
        -activebackground => $color,
        -activeforeground => 'White',
        -command => sub { $mainFrame->setPalette($color); }
    )->pack(-side => 'left', -fill => 'both', -expand => 'no');
}
$colors->pack(-side => 'left', -fill => 'both', -expand => 'no');

my $restart = $bottomFrame->Button( -text => "restart",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&game_restart],);
my $quit = $bottomFrame->Button( -text => "quit program",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => sub{exit},);
my $kugelSets = $bottomFrame->Button( -text => "Kugel Sets",
                                      -pady => '0',
                                      -padx => '0',
                                      -command => [\&kugel_choice],);
my $dimensions = $bottomFrame->Button( -text => "Dimensions",
                                       -pady => '0',
                                       -padx => '0',
                                       -command => [\&set_dimensions],);
my $demo = $bottomFrame->Button( -text => "Demo",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&demo],);
my $undo = $bottomFrame->Button( -text => "Undo",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&undo_one_move],);
my $redo = $bottomFrame->Button( -text => "Redo",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&redo_one_move],);
$kugelSets->pack( -side => 'left', -fill => 'x', -expand => 'no');
$dimensions->pack( -side => 'left', -fill => 'x', -expand => 'no');
$demo->pack( -side => 'left', -fill => 'x', -expand => 'no');
$undo->pack( -side => 'left', -fill => 'x', -expand => 'no');
$redo->pack( -side => 'left', -fill => 'x', -expand => 'no');

$restart->pack( -side => 'left', -fill => 'x', -expand => 'no');
$quit->pack( -side => 'left', -fill => 'x', -expand => 'no');

#
# menu
#
my $menuFrame = $mainFrame->Frame(-relief => 'raised', -borderwidth => 2);

#
# file menu
#
my $fileMenu = $menuFrame->Menubutton(-text => 'File', -underline => 0);
$fileMenu->command(-label => 'Hall Of Fame', -command => [\&show_hall_of_fame]);
$fileMenu->command(-label => 'Restart', -command => [\&game_restart] );
$fileMenu->command(-label => 'Undo', -command => [\&undo_one_move] );
$fileMenu->command(-label => 'Redo', -command => [\&redo_one_move] );
$fileMenu->command(-label => 'Save', -command => [\&save_game_to_file] );
$fileMenu->command(-label => 'Load', -command => [\&load_game_from_file] );
$fileMenu->command(-label => 'Quit', -command => sub{exit} );
$fileMenu->pack(-side => 'left');

#
# options menu
#
my $optionsMenu = $menuFrame->Menubutton(-text => 'Options', -underline => 0);
$optionsMenu->command(-label => 'Kugel Choice', -command=>[\&kugel_choice]);
$optionsMenu->command(-label => 'Dimensions', -command=>[\&set_dimensions]);
$optionsMenu->checkbutton(
    -label    => 'Allow skip through',
    -variable => \$allowSkip,
    -command  => sub { 
                     &get_scores; 
                     if($verbose) { print "allowSkip = $allowSkip\n"; }
                 } 
);

$optionsMenu->pack(-side => 'left');

#
# view menu
#
my $viewMenu = $menuFrame->Menubutton(-text => 'View', -underline => 0);
$viewMenu->checkbutton(
    -label => 'landscape',
    -variable => \$landscape,
    -command => sub {
                    if($verbose) { print "landscape = $landscape\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Bottom Shortcuts',
    -variable => \$showShortcuts,
    -command => sub {
                    if($verbose) { print "showShortcuts = $showShortcuts\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Score Frame',
    -variable => \$showInfo,
    -command => sub {
                    if($verbose) { print "showInfo = $showInfo\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Next Kugel',
    -variable => \$showNext,
    command => sub {
                    if($verbose) { print "showNext = $showNext\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Do Animation',
    -variable => \$doDeleteAnimation,
    command => sub {
                    if($verbose) { print "doDeleteAnimation = $doDeleteAnimation\n"; }
                }
);
$viewMenu->pack(-side => 'left');

#
# colors menu
#
my $colorsMenu = $menuFrame->Menubutton(-text => 'Colors', -underline => 0);
foreach my $color (@ApplColors) {
    $colorsMenu->command(-label => $color, 
                          -command => sub { $mainFrame->setPalette($color); },
                          -background => $color);
}
$colorsMenu->pack(-side => 'left');

#
# help menu
#
my $helpMenu = $menuFrame->Menubutton(-text => 'Help', -underline => 0);
$helpMenu->command(-label => 'How to play', -command => [\&help] );
$helpMenu->command(-label => 'Run Demo', -command => [\&demo] );
$helpMenu->command(-label => 'About', -command => [\&help_about] );
$helpMenu->pack(-side => 'right');

#
# the actual table (game itself) ...
#
for( my $j=0; $j<$boardSizeY; $j++ ) {  # make each row a different frame
    $row[$j]=$board->Frame();
}

for( my $i=0; $i<$boardSizeX; $i++ ) {
    for( my $j=0; $j<$boardSizeY; $j++ ) {
        my $position=$i+$j*$boardSizeX;
        $square[$position]=$row[$j]
            ->Button(-image => $currentSet[0],
                     -padx => '0',
                     -pady => '0'  );

        $square[$position]->bind('<1>' => [\&square_click, $position] );
        $square[$position]->bind('<Leave>' => [\&square_leave, $position] );
        $square[$position]
            ->bind('<ButtonRelease>' => [\&square_leave, $position] );
        
        $square[$position]->pack(-side => 'left');
    }
}

for( my $j=0; $j<$boardSizeY; $j++ ) {
    $row[$j]->pack( -side => 'top');
}

#
# if the higth of the screen is less than about 1000 pixels, the window
# in "portrait" layout will be higher than a screen itself. to avoid this
# situation where the window does not fit the screen (user can't see the 
# whole thing...) i'll activate the "landscape" option if the available 
# screen higth is not enough to fit the window. 
#
# both "landscape" and "portrait" layout shall still be available from 
# the menu. i could disable this for smaller screen size, but i find that 
# would be confusing confusing. 
#

if( $mainFrame->screenheight() < 1000 ) {
    if($verbose) { print "Screen height = ".$mainFrame->screenheight()
                         ." < 1000  - use 'landscape' layout\n"; }
    $landscape = 1;
}

if( $] < 5.005 ) { 
    show_error( "\nThis program has been tested \n".
                "with Perl version 5.00502\n\n".
                "You might have problems trying\n".
                "to use it with older version\n\n".
                "It might be healthy to upgrade\n".
                "your Perl installation anyway."
              )
}
if( $tkVersion < 800.012 ) {
    show_error( "\n\n\n".
                "This program has been tested\n".
                "with Tk module release 800.012\n".
                "You might have problems trying\n".
                "to use it with older version\n"
              )
}

if( ! -w $scoreFile ) {
    show_error( "\nFile: $scoreFile\n".
                "not writable for user ID $<\n\n".
                "Change permissions for the \n".
                "scorefile, if you wish to \n".
                "be able to enter the high \n".
                "score list "
              )
}

# pack the board and start the game
&re_pack;
&pick_random;
&add_kugel;

# take the very first snapshot of the game
&take_snapshot;

# everything is set to go - start the interactive game
MainLoop;

sub square_click {
    # bind will pass the Tk object as a first 
    # argument and i don't need it.
    shift;
    my $position = shift;
    if( defined( $selected )) {
        if( $square[$position]->cget(-image) == $currentSet[0]  &&  # move (?)
            path_exists($selected,$position)) {

            # if we get here, it means user has moved the kugel, or, to be
            # precise, it means user has selected to move the kugel from
            # one position to an empty square.
            my $saveSelectedImage = $square[$selected]->cget(-image);
            for( my $i=$#currentSet*$doMoveAnimation; $i>=0; $i-- ) { 
                $square[$position]->configure(-image => $currentSet[$i],
                                              -relief => 'raised' );
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
                select(undef, undef, undef, $animationSpeed);
                $square[$position]->configure(-image => $currentSet[0],
                                              -relief => 'raised' );
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
            for( my $i=$#currentSet*$doMoveAnimation; $i>=0; $i-- ) { 
                $square[$selected]->configure(-image => $currentSet[$i],
                                              -relief => 'raised' );
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
                select(undef, undef, undef, $animationSpeed);
                $square[$selected]->configure(-image => $currentSet[0],
                                              -relief => 'raised' );
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
            $square[$position]->configure(
                # -image => $square[$selected]->cget(-image),
                -image  => $saveSelectedImage,
                -relief => 'raised' );
            $square[$selected]->configure(-image => $currentSet[0],
                                          -relief => 'raised' );
            $selected=undef;
            if( my $numDeleted = &delete_kugel($position) ) {
                $score += $numDeleted;
                $scoreLabel->configure(
                    -text => "Score: ".(sprintf '%04d',$score));
                $filled -= $numDeleted;

                $numFilledLabel->configure(
                    -text => (sprintf '%02d',
                             int($filled/($boardSizeX*$boardSizeY)*100))
                             ."\% Filled");
                #
                # empty label no longer used, but it might be in the future
                #
                $numEmptyLabel->configure(
                    -text => "Empty: ".(sprintf '%02d',
                                       ($boardSizeX*$boardSizeY-$filled))
                    ." (".(sprintf '%02d',
                           int(100*($boardSizeX*$boardSizeY-$filled)
                          /($boardSizeX*$boardSizeY)))."\%)");
                #
                # title of main window
                #
                $mainFrame->configure( -title => "Kugel Score: $score");
            }
            else {
                &add_kugel;
            }

            # After user has moved a kugel, and after (perhaps) few were
            # deleted from the screen, take the snapshot of the new position
            &take_snapshot;
        }
        else {
            if( $square[$position]->cget(-image) != $currentSet[0] ) {
                $square[$selected]->configure(-relief => 'raised' );
                $square[$position]->configure(-relief => 'sunken' );
                $selected=$position;  # new select
            }
        }
    }
    elsif( $square[$position]->cget(-image) != $currentSet[0]) {  # new select
        $square[$position]->configure(-relief => 'sunken' );
        $selected=$position;
    }
}

sub square_leave {
    # bind will pass the Tk object as a first
    # argument and i don't need it.
    shift;
    my $position = shift;

    # let the selected square stay 'suunken'
    if( defined($selected) && $selected==$position ) {
        $square[$position]->configure(-relief => 'sunken' );
    }
}

sub add_kugel {
    if( @topKugel + $filled >= $boardSizeX*$boardSizeY ) {  # no space left
        &game_over;
    }

    foreach my $kugel (@topKugel) {
        my $position ;
        do { 
            $position = int rand $boardSizeX*$boardSizeY; 
        } until ( $square[$position]->cget(-image) == $currentSet[0] );
        $square[$position]->configure(-image => $kugel->cget(-image) );
        if( my $numDeleted = &delete_kugel($position) ) {
            $filled -= $numDeleted;
            $numFilledLabel->configure(
                -text => (sprintf '%02d',
                         int($filled/($boardSizeX*$boardSizeY)*100))
                         ."\% Filled");
            $numEmptyLabel->configure(
                -text => "Empty: ".(sprintf '%02d',
                                   ($boardSizeX*$boardSizeY-$filled))." ("
                .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
                      /($boardSizeX*$boardSizeY)))."\%)");
        }

    }
    $filled += @topKugel ;  # add to number of kugels displayed
    $numFilledLabel->configure(-text => (sprintf '%02d',
                                int($filled/($boardSizeX*$boardSizeY)*100))
                               ."\% Filled");
    $numEmptyLabel->configure(
        -text => "Empty: ".(sprintf '%02d',($boardSizeX*$boardSizeY-$filled))
        ." ("
        .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
              /($boardSizeX*$boardSizeY)))."\%)");
    &pick_random;
}

sub pick_random {
    foreach my $kugel (@topKugel) {  # delete the old ones
        $kugel->packForget();
    }
    @topKugel = () ;  # i hope garbage collection will stop the leak
    my $kugelsPicked =0 ;
    while( $kugelsPicked < $addKugel ) {
        push @topKugel, 
             $nextFrame->Label(-image=>$currentSet[(int rand $setOf)+1]);
        $kugelsPicked++;
    }
    &pack_next_kugel;
}

sub delete_kugel {
    my $position = shift;

    my $returnScore =0 ;

    my %numberOfKugels;  # number of kugels on each side

    $numberOfKugels{left}        = 0 ; # this init might not be necessary
    $numberOfKugels{right}       = 0 ;
    $numberOfKugels{above}       = 0 ;
    $numberOfKugels{below}       = 0 ;
    $numberOfKugels{aboveleft}   = 0 ;
    $numberOfKugels{aboveright}  = 0 ;
    $numberOfKugels{belowleft}   = 0 ;
    $numberOfKugels{belowright}  = 0 ;

    #
    # left
    #
    for( my $tempPos = $position -1 ;
         ($tempPos+1)%$boardSizeX  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos--) {
        $numberOfKugels{left}++
    }

    #
    # right
    #
    for( my $tempPos = $position +1 ;
         $tempPos%$boardSizeX  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos++) {
        $numberOfKugels{right}++
    }

    #
    # above
    #
    for( my $tempPos = $position -$boardSizeX ;
         $tempPos>=0  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX ) {
        $numberOfKugels{above}++
    }

    #
    # below
    #
    for( my $tempPos = $position +$boardSizeX ;
         $tempPos < $boardSizeX*$boardSizeY  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX ) {
        $numberOfKugels{below}++
    }

    #
    # aboveleft
    #
    for( my $tempPos = $position -$boardSizeX -1 ;
         $tempPos>=0  && ($tempPos+1)%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX+1 ) {
        $numberOfKugels{aboveleft}++
    }

    #
    # aboveright
    #
    for( my $tempPos = $position -$boardSizeX +1 ;
         $tempPos>=0  && $tempPos%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX-1 ) {
        $numberOfKugels{aboveright}++
    }

    #
    # belowleft
    #
    for( my $tempPos = $position +$boardSizeX-1 ;
         $tempPos < $boardSizeX*$boardSizeY  && ($tempPos+1)%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX-1 ) {
        $numberOfKugels{belowleft}++
    }

    #
    # belowright
    #
    for( my $tempPos = $position +$boardSizeX+1 ;
         $tempPos < $boardSizeX*$boardSizeY  && $tempPos%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX+1 ) {
        $numberOfKugels{belowright}++
    }

    if( $numberOfKugels{right} + $numberOfKugels{left} >= $delKugel ) {
        for( my $count=1 ; $count<=$numberOfKugels{left}; $count++) {
          for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                $square[$position-$count]->configure(-image => $currentSet[$i],
                                                     -relief => 'raised' );
              # let Tk update the screen, then sleep few miliseconds 
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
              select(undef, undef, undef, $animationSpeed);
                  $square[$position-$count]->configure(-image => $currentSet[0],
                                                     -relief => 'raised' );
              # let Tk update the screen
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        for( my $count=1 ; $count<=$numberOfKugels{right}; $count++) {
          for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                $square[$position+$count]->configure(-image => $currentSet[$i],
                                                     -relief => 'raised' );
                # let Tk update the screen, then sleep few miliseconds 
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
                select(undef, undef, undef, $animationSpeed);
                    $square[$position+$count]->configure(-image => $currentSet[0],
                                                         -relief => 'raised' );
                # let Tk update the screen
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore += $numberOfKugels{right} + $numberOfKugels{left} ;
    }
    if( $numberOfKugels{above} + $numberOfKugels{below} >= $delKugel ) {
        for( my $count=1 ; $count<=$numberOfKugels{above}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                  $square[$position-$count*$boardSizeX]->configure(
                                                       -image => $currentSet[$i],
                                                       -relief => 'raised' );
                  # let Tk update the screen, then sleep few miliseconds 
                  for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
                  select(undef, undef, undef, $animationSpeed);
                      $square[$position-$count*$boardSizeX]->configure(
                                                         -image => $currentSet[0],
                                                         -relief => 'raised' );
                  # let Tk update the screen
                  for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
              }
          }
          for( my $count=1 ; $count<=$numberOfKugels{below}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                  $square[$position+$count*$boardSizeX]->configure(
                                                       -image => $currentSet[$i],
                                                       -relief => 'raised' );
              # let Tk update the screen, then sleep few miliseconds 
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
              select(undef, undef, undef, $animationSpeed);
                  $square[$position+$count*$boardSizeX]->configure(
                                                       -image => $currentSet[0],
                                                       -relief => 'raised' );
              # let Tk update the screen
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore += $numberOfKugels{above} + $numberOfKugels{below} ;
    }
    if( $numberOfKugels{aboveleft} + $numberOfKugels{belowright} >=$delKugel){
        for( my $count=1 ; $count<=$numberOfKugels{aboveleft}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                    $square[$position-$count*($boardSizeX+1)]->configure(
                                                         -image => $currentSet[$i],
                                                         -relief => 'raised' );
                # let Tk update the screen, then sleep few miliseconds 
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
                select(undef, undef, undef, $animationSpeed);
                    $square[$position-$count*($boardSizeX+1)]->configure(
                                                         -image => $currentSet[0],
                                                       -relief => 'raised' );
                # let Tk update the screen
                for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        for( my $count=1 ; $count<=$numberOfKugels{belowright}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                  $square[$position+$count*($boardSizeX+1)]->configure(
                                                       -image => $currentSet[$i],
                                                       -relief => 'raised' );
            # let Tk update the screen, then sleep few miliseconds 
            for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            select(undef, undef, undef, $animationSpeed);
                  $square[$position+$count*($boardSizeX+1)]->configure(
                                                       -image => $currentSet[0],
                                                       -relief => 'raised' );
            # let Tk update the screen
            for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore +=$numberOfKugels{aboveleft}+$numberOfKugels{belowright};
    }
    if( $numberOfKugels{aboveright} + $numberOfKugels{belowleft} >=$delKugel){
        for( my $count=1 ; $count<=$numberOfKugels{aboveright}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                  $square[$position-$count*($boardSizeX-1)]->configure(
                                                       -image => $currentSet[$i],
                                                       -relief => 'raised' );
              # let Tk update the screen, then sleep few miliseconds 
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
              select(undef, undef, undef, $animationSpeed);
                  $square[$position-$count*($boardSizeX-1)]->configure(
                                                       -image => $currentSet[0],
                                                       -relief => 'raised' );
              # let Tk update the screen
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        for( my $count=1 ; $count<=$numberOfKugels{belowleft}; $count++) {
            for( my $i=$#currentSet*$doDeleteAnimation; $i>=0; $i-- ) {
                  $square[$position+$count*($boardSizeX-1)]->configure(
                                                       -image => $currentSet[$i],
                                                       -relief => 'raised' );
              # let Tk update the screen, then sleep few miliseconds 
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
              select(undef, undef, undef, $animationSpeed);
                  $square[$position+$count*($boardSizeX-1)]->configure(
                                                       -image => $currentSet[0],
                                                       -relief => 'raised' );
              # let Tk update the screen
              for(my $j=0; $j<10; $j++) { DoOneEvent(Tk::DONT_WAIT); }
            }
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore +=$numberOfKugels{aboveright}+$numberOfKugels{belowleft};
    }

    if( $returnScore == 0 ) { return $returnScore; }
    else { return $returnScore+1; }
}

sub kugel_choice {
    my $kugelChoiceDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                                -title => "Kugel Choice",
                                                -padx => '0');

    my %frame ;

    foreach my $dir (@dir) {
        $frame{$dir}=$kugelChoiceDialog->Frame(-relief =>'raised', 
                                               -borderwidth =>2);
        if( $#{$photo{$dir}} >= $setOf ) {
            foreach my $photo (@{$photo{$dir}}) {
                $frame{$dir}->Button(-image => $photo,
                                     -command => [\&new_kugel_choice, $dir] 
                              )->pack(-side => 'left', 
                                      -fill => 'x', 
                                      -expand => 'yes');
            }
        }
        else {
            foreach my $photo (@{$photo{$dir}}) {
                $frame{$dir}->Button(-image => $photo,
                                     -state => 'disabled',
                              )->pack(-side => 'left',
                                      -fill => 'x',
                                      -expand => 'yes');
            }
        }
    }

    foreach my $dir (@dir) { $frame{$dir}->pack(); }

    $kugelChoiceDialog->Show();
}

sub new_kugel_choice {
    my $choice=shift;
    my @newPhoto;

    foreach my $dir (@dir) {
        if( $choice eq $dir ) {
            @newPhoto = @{$photo{$dir}}
        }
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position = $i+$j*$boardSizeX;
            for( my $p=0; $p<=$#currentSet; $p++) {
                if( $square[$position]->cget(-image) == $currentSet[$p] ) {
                    if( defined $newPhoto[$p] ) {
                        $square[$position]->configure(-image => $newPhoto[$p]);
                    }
                    else { 
                        # only coding error can bring me here
                        if($verbose) { print "not enough kugels in new set\n";}
                        @currentSet=@newPhoto;
                        $setOf = $#currentSet;
                        if($verbose) { print "setOf = $setOf\n"; }
                        &full_restart;  # not enough kugels in new set
                    }
                }
            }
        }
    }
    for( my $i=0; $i<$addKugel ; $i++ ){
        for( my $p=0; $p<=$#currentSet; $p++) {
            if( $topKugel[$i]->cget(-image) == $currentSet[$p] ) {
                if( defined $newPhoto[$p] ) {
                    $topKugel[$i]->configure(-image => $newPhoto[$p]);
                }
                else {
                    # only coding error can bring me here
                    if($verbose) { print "not enough kugels in new set (2)\n"; }
                    @currentSet=@newPhoto;
                    $setOf = $#currentSet;
                    if($verbose) { print "setOf = $setOf\n"; }
                    &full_restart;  # not enough kugels in new set
                }
            }
        }
    }
    @currentSet=@newPhoto;
}

sub help_about {
    my $helpAboutDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                              -title => "Help About",
                                              -padx => '0');
    my $frame1=$helpAboutDialog->Frame(-relief =>'flat', -borderwidth =>2);

    my $text1 = $frame1->Text(-relief => 'flat',
                              -width => 42, 
                              -height => 10,
                              -cursor => 'top_left_arrow',
                              -font => $helpFont,
                              -setgrid => 'yes' );
    $text1->insert('end',"\n This program was made by Vlatko Primorac\n");
    $text1->insert('end',"\n vlatko\@vlptel.com\n");
    $text1->insert('end',"\n You may freely distribute it under \n");
    $text1->insert('end',"\n GNU general public license\n");

    $text1->pack(-side => 'left', -fill => 'x', -expand => 'yes');

    $frame1->pack();
    $helpAboutDialog->Show();

}

sub game_over {

    if( !defined($scores[9]) || $score>substr($scores[9],0,4) ) {
        &write_score;
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0],
                                          -relief => 'raised',     );
        }
    }
    $score=0;
    $didSkip=0;
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $filled=0;
    &pick_random;
}

sub game_restart {
    if($verbose) { print "sub game_restart\n"; }
    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0],
                                          -relief => 'raised',     );
        }
    }
    $score=0;
    $didSkip=0;
    $filled=0;
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    &pick_random;

    # there is a remote possibility this will result in an
    # empty board, but that's likely only on "extreme" sets
    # like delete by 2 and set of 1. in that case, the game
    # should be re-restarted.
    while( $filled==0 ) {
        if($verbose) { print "populating the board, calling add_kugel\n"; }
        &add_kugel;
    }
    $mainFrame->configure( -title => "Kugel");

    # take a new snapshot after restart
    &take_snapshot;
}

sub full_restart {
    if($verbose) { print "sub full_restart\n"; }

    $board->packForget();
    foreach my $row    (@row)    { $row->packForget(); }
    foreach my $square (@square) { $square->packForget(); }

    for( my $j=0; $j<$boardSizeY; $j++ ) {  # make each row a different frame
        $row[$j]=$board->Frame();
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]=$row[$j]
                ->Button(-image => $currentSet[0],
                         -padx => '0',
                         -pady => '0',
                         -relief => 'raised',
                         # -command => [\&square_click, $position]
            );

            # "bind" here 
            $square[$position]->bind('<1>' => [\&square_click, $position]);
            $square[$position]->bind('<Leave>' => [\&square_leave, $position]);
            $square[$position]
                ->bind('<ButtonRelease>' => [\&square_leave, $position]);

            $square[$position]->pack(-side => 'left');
        }
    }

    for( my $j=0; $j<$boardSizeY; $j++ ) {
        $row[$j]->pack( -side => 'top');
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0]);
        }
    }
    $score=0;
    $didSkip=0;
    $filled=0;

    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $setOfLabel->configure( -text => "Set Of $setOf ");
    $addLabel  ->configure( -text => "Add by $addKugel ");
    $delLabel  ->configure( -text => "Delete by ".($delKugel+1) );
    # $numFilledLabel->configure();

    &pick_random;

    # there is a remote possibility this will result in an
    # empty board, but that's likely only on "extreme" sets
    # like delete by 2 and set of 1. in that case, the game 
    # should be re-restarted.
    while( $filled==0 ) {
        if($verbose) { print "populating the board, calling add_kugel\n"; }
        &add_kugel;
    }
    $board->pack();
    $mainFrame->configure( -title => "Kugel");

    # take a snapshot after restart
    &take_snapshot;
}

#
# determine if the clear path exists between two positions
#
sub path_exists {

    my $position1 = shift;
    my $position2 = shift;

    if($verbose) { print "sub path_exists($position1,$position2)\n"; }

    my @board;
    my @animate;  # squares to animate, if needed

    my $longest = $boardSizeX*$boardSizeY; # longest possible path (?)

    for( my $i=0; $i<$boardSizeX*$boardSizeY; $i++ ) { $board[$i]=$longest+1; }

    $board[$position1] =1 ;


    for( my $i=1; $i<$longest; $i++ ) {
        for( my $position=0; $position<$boardSizeX*$boardSizeY; $position++ ) {
            if( $board[$position] == $i ) {
                if($verbose) { print "position=$position reached\n"; }
                #
                # left
                #
                if( $position%$boardSizeX ) {
                    if( $position-1 == $position2 ) { goto bingo; }
                    if( $square[$position-1]->cget(-image) == $currentSet[0] &&
                        $board[$position-1] > $i+1 ) {
                        $board[$position-1] = $i+1;
                    }
                }
                #
                # right
                #
                if( ($position+1)%$boardSizeX ) {
                    if( $position+1 == $position2 ) { goto bingo; }
                    if( $square[$position+1]->cget(-image) == $currentSet[0] &&
                        $board[$position+1] > $i+1 ) {
                        $board[$position+1] = $i+1;
                    }
                }
                #
                # above
                #
                if( $position - $boardSizeX  >= 0 ) {
                    if( $position-$boardSizeX == $position2 ) { goto bingo; }
                    if( $square[$position-$boardSizeX]
                        ->cget(-image)==$currentSet[0]  &&
                        $board[$position-$boardSizeX] > $i+1 ) {
                        $board[$position-$boardSizeX] = $i+1;
                    }
                }
                #
                # below
                #
                if( $position +$boardSizeX < $boardSizeX*$boardSizeY ) {
                    if( $position+$boardSizeX == $position2 ) { goto bingo; }
                    if( $square[$position+$boardSizeX]
                        ->cget(-image)==$currentSet[0]  &&
                        $board[$position+$boardSizeX] > $i+1 ) {
                        $board[$position+$boardSizeX] = $i+1;
                    }
                }
            }
        }
    }
bingo:
    if( $position2%$boardSizeX &&                             # left
        $board[$position2-1] <= $longest) {
        if($verbose) { print "Path length = $board[$position2-1]\n"; }
        return $board[$position2-1];
    } 
    if( ($position2+1)%$boardSizeX &&                         # right
        $board[$position2+1] <= $longest) {
        if($verbose) { print "Path length = $board[$position2+1]\n"; }
        return $board[$position2+1];
    }
    if( $position2-$boardSizeX >=0 &&                          # above
        $board[$position2-$boardSizeX] <=$longest) { 
        if($verbose) { print "Path length = $board[$position2-$boardSizeX]\n"; }
        return $board[$position2-$boardSizeX];
    }
    if( $position2+$boardSizeX < $boardSizeX*$boardSizeY &&   # below
        $board[$position2+$boardSizeX] <= $longest) {
        if($verbose) { print "Path length = $board[$position2+$boardSizeX]\n"; }
        return $board[$position2+$boardSizeX];
    }

    if($verbose) { print "Path not found\n"; }

    if( $allowSkip ) { 
        if($verbose) { print "Skip allowed: allowSkip=$allowSkip\n"; }
        $didSkip=1;
        return 1; 
    }

    return # return undef (false)
}

sub set_dimensions {
    if($verbose) { print "sub set_dimensions\n"; }
    my $dimensionsDialog = $mainFrame->Dialog( -buttons => ["OK", "Cancel"], 
                                               -title => "Set Dimensions",
                                               -padx => '0');

    my $oldSizeX = $boardSizeX;
    my $oldSizeY = $boardSizeY;
    my $oldAdd   = $addKugel;
    my $oldDel   = $delKugel;
    my $oldSetOf = $setOf;

    my $scaleYtext = $dimensionsDialog->Text(-relief => 'flat',
                                             -width => 5,
                                             -height => 1,
                                             -font => $bigFont,
                                             -cursor => 'top_left_arrow',
                                             -setgrid => 'yes');
    $scaleYtext->insert('end','Rows:');

    my $scaleY = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 4,
        -to     => 16,
        -length => 200,
        -tickinterval => 2,
        -command => sub { $boardSizeY=shift; }
    );
    $scaleY->set($boardSizeY);

    $scaleYtext->pack();
    $scaleY->pack();

    my $scaleXtext = $dimensionsDialog->Text(-relief => 'flat',
                                             -width => 8,
                                             -height => 1,
                                             -font => $bigFont,
                                             -cursor => 'top_left_arrow',
                                             -setgrid => 'yes');
    $scaleXtext->insert('end','Columns:');

    my $scaleX = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 4,
        -to     => 16,
        -length => 200,
        -tickinterval => 2,
        -command => sub { $boardSizeX=shift; }
    );
    $scaleX->set($boardSizeX);

    $scaleXtext->pack();
    $scaleX->pack();

    my $addText = $dimensionsDialog->Text(-relief => 'flat',
                                          -width => 7,
                                          -height => 1,
                                          -font => $bigFont,
                                          -cursor => 'top_left_arrow',
                                          -setgrid => 'yes');
    $addText->insert('end','Add by:');

    my $add = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 2,
        -to     => 6,
        -length => 200,
        -tickinterval => 1,
        -command => sub { $addKugel=shift; }
    );
    $add->set($addKugel);

    $addText->pack();
    $add->pack();

    my $delText = $dimensionsDialog->Text(-relief => 'flat',
                                          -width => 10,
                                          -height => 1,
                                          -font => $bigFont,
                                          -cursor => 'top_left_arrow',
                                          -setgrid => 'yes');
    $delText->insert('end','Delete by:');

    my $del = $dimensionsDialog->Scale( 
        -orient => 'horizontal',
        -from   => 2,
        -to     => 6,
        -length => 200,
        -tickinterval => 1,
        -command => sub { $delKugel=(shift)-1; });
    $del->set($delKugel+1);

    $delText->pack();
    $del->pack();

    my $setOfText = $dimensionsDialog->Text(-relief => 'flat',
                                            -width => 10,
                                            -height => 1,
                                            -font => $bigFont,
                                            -cursor => 'top_left_arrow',
                                            -setgrid => 'yes');
    $setOfText->insert('end','Use set of:');

    my $set = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 1,
        -to     => $#currentSet,  # let it scale to the number available
        -length => 200,
        -tickinterval => 1,
        -command => sub { $setOf=shift; }
    );
    $set->set($setOf);

    $setOfText->pack();
    $set->pack();

    my $button_pressed = $dimensionsDialog->Show();

    if($verbose) { print "Button pressed: ".$button_pressed."\n"; }

    if( $button_pressed eq "Cancel" ) {
        $boardSizeX = $oldSizeX;
      $boardSizeY = $oldSizeY;
      $addKugel   = $oldAdd;
      $delKugel   = $oldDel;
      $setOf      = $oldSetOf;

      if($verbose) { print "Original dimensions restored.\n"; } 
      return;
    }

    if( $boardSizeX != $oldSizeX || $boardSizeY != $oldSizeY ||
        $oldAdd != $addKugel     || $oldDel != $delKugel     ||
        $oldSetOf != $setOf                                     ) {

        if($verbose) { print "boardSizeX = $boardSizeX \n"; }
        if($verbose) { print "boardSizeY = $boardSizeY \n"; }
        if($verbose) { print "addKugel = $addKugel \n"; }
        if($verbose) { print "delKugel = $delKugel \n"; }
        if($verbose) { print "setOf = $setOf \n"; }

        # delete history
        @gameHistory = ();
        $gameHistoryPosition = -1;
        if($verbose) { print "History deleted.\n"; }

      if($verbose) { print "Calling full_restart after dimensions have changed.\n"; }
        &full_restart; # restart the game
        &get_scores;   # get the scores for the new settings
    } else {
        if($verbose) { print "Dimensions unchanged. Nothing to do.\n"; }
    }
}

sub get_scores {
    my $SCORES;       # file handler
    open(SCORES, $scoreFile ) || die "could not open file $scoreFile: $!"  ;

    @scores=();

    while( defined( my $line = <SCORES> ) ) {
        if( $line !~ /^Rows:\s+/ ) { next; }  # skip junk...
        if( $line =~ /^Rows: (\d+) Cols: (\d+) Add: (\d+) Del: (\d+) Set: (\d+) Skip: ([0,1]) Score: (\d{4}) Name: (\w.*)/ ) {
            my ($rows, $cols, $add, $del, $set, $skip, $score, $name) = 
               ($1,    $2,    $3,   $4,   $5,   $6,    $7,     $8);
            if( $rows eq $boardSizeY &&
                $cols eq $boardSizeX &&
                $add  == $addKugel   &&
                $del  == $delKugel+1 &&
                $set  == $setOf      &&
                $skip == $allowSkip    ) {
                push @scores, "$score Name: $name";
            }
        }
    }

    close SCORES;

    # reversed order, sort lexically
    @scores = sort {$b cmp $a} @scores;

    if( $verbose ) {
        print "Scores: \n";
        foreach my $scoreline (@scores) { print "$scoreline\n"; }
    }

    if( defined( $scores[0] )) {
        $scores[0] =~ /^(\d{4})/;
        $highestScore = $1;
        $scores[0] =~ /Name: (\w.*)/;
        $champion = $1;
    }
    else {
        $highestScore = "0000";
        $champion = "1st time played";
    }

    if( defined($highScoreLabel) ) {  # rewrite the highest score
        $highScoreLabel->configure( 
            -text => "$champion: ".(sprintf '%04d',$highestScore) );
    }

    if( $verbose ) { print "highest score: $highestScore\n"; }
    if( $verbose ) { print "champion: $champion\n"; }

}

sub write_score {
    my $SCORES;                                 # file handler
    my $name="enter your name here...";         # user input name

    my $scoreDialog = $mainFrame->Dialog(
        -buttons => undef, 
        -title => "Congratulations!!! You have made it to the hall of fame.",
        -padx => '0');

    my $frame1=$scoreDialog->Frame(-relief =>'raised', -borderwidth =>2);
    my $frame2=$scoreDialog->Frame(-relief =>'raised', -borderwidth =>2);

    my $scoreText = $scoreDialog->Text(-relief => 'flat',
                                       -width => 40,
                                       -height => 5,
                                       -font => $scoreFont,
                                       -cursor => 'top_left_arrow',
                                       -setgrid => 'yes');

    $scoreText->insert('end', "\n           Congratulations !!!\n\n");
    $scoreText->insert('end', " You have made it to the hall of fame.\n");
    $scoreText->pack();

    my $entry = $frame2->LabEntry( -relief => 'sunken', 
                                   -width => 40,
                                   -textvariable => \$name);
    $entry->grid(-sticky => 'ew');
    $entry->focus();

    $frame1->pack();
    $frame2->pack();
    $scoreDialog->Show();

    $score = sprintf '%04d',$score;  # format to 0000

    if( $name eq "enter your name here..." ) { return; }

    open(SCORES, ">>$scoreFile" ) || warn "could not open file $scoreFile: $!";

    print SCORES "Rows: $boardSizeY ";
    print SCORES "Cols: $boardSizeX ";
    print SCORES "Add: $addKugel ";
    print SCORES "Del: ".($delKugel+1)." ";
    print SCORES "Set: ".($setOf)." ";
    print SCORES "Skip: $didSkip ";
    print SCORES "Score: $score Name: $name\n";
    close SCORES;

    if( $verbose) {
        print "Rows: $boardSizeY ";
        print "Cols: $boardSizeX ";
        print "Add: $addKugel ";
        print "Del: ".($delKugel+1)." ";
        print "Skip: $didSkip ";
        print "Score: $score Name: $name\n"; 
    }

    &get_scores;  # reload the scores
}

sub show_hall_of_fame {
    my $fameDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                         -title => "Hall of Fame",
                                         -padx => '0');

    my $fameText = $fameDialog->Text(-relief => 'flat',
                                     -width => 25,
                                     -height => 11,
                                     -font => $scoreFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $fameText->insert('end', "\n       Hall of Fame: \n\n");

    foreach my $scoreline (@scores) {
        $scoreline =~ /^(\d{4}) Name: (\w.*)/;
        $fameText->insert('end', " $1 $2 \n");
    }

    $fameText->pack();
    $fameDialog->Show();
}

#
# error dialog
#
sub show_error {
    my $text = shift;
    my $errorDialog = $mainFrame->Dialog( -buttons => ['Close'],
                                         -title => "Error Message",
                                         -padx => '0');

    my $errorText = $errorDialog->Text(-relief => 'flat',
                                       -width => 30,
                                       -height => 10,
                                       -font => $scoreFont,
                                       -cursor => 'top_left_arrow',
                                       -setgrid => 'yes');

    $errorText->insert('end', $text);

    $errorText->pack();
    $errorDialog->Show();
}

sub help {
    if($verbose) { print "sub help\n"; }
    my $helpDialog = $mainFrame->Dialog( -buttons => ['Close'],
                                         -title => "How to play the game",
                                         -padx => '0');

    my $helpText = $helpDialog->Text(-relief => 'flat',
                                     -width => 62,
                                     -height => 23,
                                     -font => $helpFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $helpText->insert('end', 
        "\n                       How to play it? \n\n");
    $helpText->insert('end',  "\n");
    $helpText->insert('end',  
        'This is version 2.4 of the perl/tk based  game I call "kugel",'."\n");
    $helpText->insert('end',  
        'you can  call it whatever you  please.  The game  is played by'."\n");
    $helpText->insert('end',  
        'moving  objects  on  the board  and putting  them  into  rows,'."\n");
    $helpText->insert('end',  
        'columns, diagonals.  Each time you create  a line of  the same'."\n");
    $helpText->insert('end',  
        'objects  ("kugels", balls...), the score  will  increment with'."\n");
    $helpText->insert('end',  
        'the  number  of objects  you  accomplished to line up.  In the'."\n");
    $helpText->insert('end',  
        'same time,  the lined up "kugels"  will  disappear, so you can'."\n");
    $helpText->insert('end',  
        'continue the game with more maneuver  space on the board.  Any'."\n");
    $helpText->insert('end',  
        'time  you  fail to create  a line of "kugels",  new ones  will'."\n");
    $helpText->insert('end',  
        'appear on the board. Once the board is full, the game is over.'."\n");
    $helpText->insert('end',  "\n");
    $helpText->insert('end',  
        "The latest and greatest version of the game is available at:\n");
    $helpText->insert('end',  
        "http://www.sunsite.unc.edu/pub/Linux/games/strategy/\n");
    $helpText->insert('end', "\n");
    $helpText->insert('end',  
        "Have fun. :)\n\nVlatko Primorac, 1998-2001\nvlatko\@vlptel.com\n\n");

    $helpText->pack();

    $helpDialog->Show();

}

#
# run demo dialog
#
sub demo {
    if($verbose) { print "sub demo\n"; }
    my $demoDialog = $mainFrame->Dialog( -default_button => 'Start',
                                         -buttons => ['Done'],
                                         -title => "Game Demo",
                                         -padx => '0');

    my $fameText = $demoDialog->Text(-relief => 'flat',
                                     -width => 25,
                                     -height => 3,
                                     -font => $scoreFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $fameText->insert('end', "\n       Game Demo: \n\n");
    $fameText->pack();

    my $startButton;
    my $stepButton;
    my $moveButton;

    $startButton = $demoDialog->Button( 
        -text => "Start",
        -command=> sub{ 
                       $startButton->packForget(); 
                       $moveButton->pack(); 
                       &game_restart;
                   },
        -font => $scoreFont);

    $stepButton = $demoDialog->Button(-text => "Step",
                                      -command=> sub{ 
                                                     &pick_random;
                                                     &add_kugel;
                                                     $stepButton->packForget();
                                                     $moveButton->pack();
                                                 },
                                      -font => $scoreFont);

    $moveButton = $demoDialog->Button(-text => "Move",
                                      -command=> sub{  
                                                     &auto_move;
                                                     $moveButton->packForget();
                                                     $stepButton->pack();
                                                 },
                                      -font => $scoreFont);
    $startButton->pack();

    $demoDialog->Show();

}

#
# calculate and make a move
#
sub auto_move {
    if($verbose) { print "sub auto_move\n"; }
    # righ now, i have no idea how to make
    # the best move, so i'll just pick one

    my $start; 
    my $destination;

    do{ $start = int rand ($boardSizeX*$boardSizeX) }
    until($square[$start]->cget(-image) != $currentSet[0]);

    do{ $destination = int rand ($boardSizeX*$boardSizeX) }
    until($square[$destination]->cget(-image) == $currentSet[0]);

    if( $verbose) { print "Start: $start, Destination: $destination\n" };

    $square[$destination]->configure( -image => $square[$start]->cget(-image));
        # -image $square[$start]->configure(cget(-image));

    $square[$start]->configure(-image => $currentSet[0]);
}

# 
# re-pack everything
#
sub re_pack {
    if($verbose) { print "sub re_pack\n"; }

    $board->packForget();               # packForget all
    $nextFrame->packForget();
    $menuFrame->packForget();
    $semaphoreFrame->packForget();
    $bottomFrame->packForget();
    
    $menuFrame->pack(-fill => 'x');     # and re-pack all

    if( $showInfo ) {
        $scoreFrame->packForget();
        $infoFrame->packForget();
        $highScoreLabel->packForget();
        $scoreLabel->packForget();

        if( $landscape ) {
            $highScoreLabel->pack(-side =>'top', -expand =>'yes', -fill =>'x');
            $scoreLabel    ->pack(-side =>'top', -expand =>'yes', -fill =>'x');
        } else {
            $highScoreLabel->pack(-side =>'left', -expand =>'yes',-fill =>'x');
            $scoreLabel    ->pack(-side =>'left', -expand =>'yes',-fill =>'x');
        }

        if( $landscape ) {
            $setOfLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $addLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $delLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $numFilledLabel->pack(-side =>'top', -expand =>'yes', -fill =>'x');
        } else {
            $setOfLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $addLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $delLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $numFilledLabel->pack(-side =>'left', -expand =>'yes',-fill =>'x');
        }

        $scoreFrame->pack( -side => 'top', -fill => 'x', -expand => 'yes');
        $infoFrame->pack( -side => 'bottom', -fill => 'x', -expand => 'yes');

        if( $landscape ) {
            $semaphoreFrame->pack( -side => 'right', 
                                   -fill => 'y', 
                                   -expand => 'yes');
        } else {
            $semaphoreFrame->pack( -side => 'top', 
                                   -fill => 'x', 
                                   -expand => 'yes');
        }
    }

    if( $showNext ) {
        &pack_next_kugel;
        if( $landscape ) {
            $nextFrame->pack( -side => 'left',  -expand => 'no');
        } else {
            $nextFrame->pack( -side => 'top',  -expand => 'no');
        }
    }

    if( $showShortcuts ) { 
        $bottomFrame->pack( -side => 'bottom');
    }

    $board->pack();

    if($verbose) {
        print "Screen width:  ".$mainFrame->screenwidth()."\n";
        print "Screen height: ".$mainFrame->screenheight()."\n";
        print "Window width:  ".$mainFrame->width()." (to be updated)\n";
        print "Window height: ".$mainFrame->height()." (to be updated)\n";
    }
}

#
# pack the "next" kugel set
#
sub pack_next_kugel {
    if($verbose) { print "sub pack_next_kugel\n"; }
    foreach my $kugel (@topKugel) { $kugel->packForget() }
    unless( $showNext ) { return }
    foreach my $kugel (@topKugel) {
        if( $landscape ) {
            $kugel->pack(-side => 'top', -fill => 'x', -expand => 'yes');
        } else {
            $kugel->pack(-side => 'left', -fill => 'x', -expand => 'yes');
        }    
    }
}

#
# takes the snapshot of the board and kugel set to add, with the
# score info. used by undo/redo functionality. 
# 
# @gameHistory - globally accessible table containing snapshots
#                it is populated with references to array, each array
#                containing one snapshot
# 
# $gameHistoryPosition - globally accessible pointer (integer array index)
#                        pointing to the current position used by undo/redo
#                        functionality. taking the new snapshot will always
#                        truncate the @gameHistory array to the lenght of
#                        $gameHistoryPosition and add the current snapshot 
#                        on top of it - after meking the new move, one can 
#                        no longer redo the previous move.
#
sub take_snapshot {
    if($verbose) { print "sub take_snapshot\n"; }
    #
    # populate hte game history
    #
    my @gameSnapshot;

    # remember the board position
    foreach my $position (@square) {
        push @gameSnapshot, $position->cget(-image);
    }

    # remember the top kugel set
    foreach my $position (@topKugel) { 
        push @gameSnapshot, $position->cget(-image);
    }

    # remember the current set of kugels being used

    # save the number of kugels in current set first
    push @gameSnapshot, $#currentSet;

    # save the images in the current set
    foreach my $position (@currentSet) {
        push @gameSnapshot, $position;
    }

    # remember the skip info, score, highest score, champion name
    push @gameSnapshot, $didSkip;
    push @gameSnapshot, $highestScore;
    push @gameSnapshot, $champion;
    push @gameSnapshot, $score;

    # truncate the history array to the current position, after new move
    # user can not redo previously undoned moves
    if( $verbose) { print "Game history truncated to: ".$gameHistoryPosition."\n"; }
    $#gameHistory = $gameHistoryPosition;

    # add current snapshot
    push @gameHistory, \@gameSnapshot;

    # $#gameHistory will point to the last element in the @gameHistory array
    $gameHistoryPosition = $#gameHistory;

    if( $verbose) { print "Snapshots in game history: ".@gameHistory."\n"; }
    if( $verbose) { print "Items stored in snapshot: ".@gameSnapshot."\n"; }
    if( $verbose) { print "Game history position: ".$gameHistoryPosition."\n"; }
}
#
# implement undo functionality
#
sub undo_one_move {
    if($verbose) { print "sub undo_one_move\n"; }
    # if there is only one position of the stack, it is the current one
    # so we have nothing to undo
    if( $gameHistoryPosition < 1 ) {
        if( $verbose ) { print "Nothing to undo.\n"; }
        return;
    }

    # when doing undo, we will no longer keep track of selected kugel
    $square[$selected]->configure( -relief => 'raised' );
    $selected=undef;

    # use the position preceeding the current one - decrement the history position
    $gameHistoryPosition--;

    if( $verbose) { print "Snapshots in game history: ".@gameHistory."\n"; }
    if( $verbose) { print "Game history position: ".$gameHistoryPosition."\n"; }

    # then set the position to the new top of the stack
    my @gameSnapshot = @{$gameHistory[$gameHistoryPosition]};

    if( $verbose) { print "Items found in snapshot: ".@gameSnapshot."\n"; }

    # restore the board position
    foreach my $position (@square) {
        $position->configure( -image => shift @gameSnapshot);
    }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }
    if( $verbose) { print "Top kugel #: ".@topKugel."\n"; }

    # restore the top kugel set
    foreach my $position (@topKugel) {
        $position->configure( -image => shift @gameSnapshot);
    }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # set the number in the current set
    $#currentSet = shift @gameSnapshot;

    if( $verbose) { print "Current set #: ".@currentSet."\n"; }

    # restore the current set of kugels being used
    foreach my $position (@currentSet) {
        $position = shift @gameSnapshot;
    }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # restore the skip info, score, highest score, champion name
    $didSkip      = shift @gameSnapshot;
    $highestScore = shift @gameSnapshot;
    $champion     = shift @gameSnapshot;
    $score        = shift @gameSnapshot;

    if($verbose) { print "didSkip="     .$didSkip      ."\n"; }
    if($verbose) { print "highestScore=".$highestScore ."\n"; }
    if($verbose) { print "champion="    .$champion     ."\n"; }
    if($verbose) { print "score="       .$score        ."\n"; }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # recalculate the number of squares filled
    $filled = 0;
    foreach my $position (@square) {
        unless( $position->cget(-image) == $currentSet[0] ) {
            $filled++;
        }
    }
    if( $verbose) { print "Filled squares: ".$filled."\n"; }

    # re-display the Filled button on the screen
    $numFilledLabel->configure(-text => (sprintf '%02d',
                                int($filled/($boardSizeX*$boardSizeY)*100))
                               ."\% Filled");

    # re-display score and highest score
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $highScoreLabel->configure(-text =>"$champion: ".(sprintf '%04d',$highestScore));
    $mainFrame->configure( -title => "Kugel Score: $score");
}

#
# implement redo functionality
#
sub redo_one_move {
    # if we are at the top of the stack, there is nothing to redo
    if( $gameHistoryPosition > $#gameHistory ) {
        if( $verbose ) { print "Coding error: gameHistoryPosition = ".$gameHistoryPosition.
                               "is greater than gameHistory size = ".$#gameHistory."\n"; }
        # show error to user, maybe he'll report it back to the author
        &show_error("Coding error: gameHistoryPosition = ".$gameHistoryPosition.
                    "is greater than gameHistory size = ".$#gameHistory."\n\n".
                    'Please report this error to author vlatko@vlptel.com Thanks!');
        return;
    }

    if( $gameHistoryPosition == $#gameHistory ) {
        if( $verbose ) { print "Nothing to redo.\n"; }
        return;
    }

    # when doing redo, we will no longer keep track of selected kugel
    $square[$selected]->configure( -relief => 'raised' );
    $selected=undef;

    # increment the history position
    $gameHistoryPosition++;

    if( $verbose) { print "Snapshots in game history: ".@gameHistory."\n"; }
    if( $verbose) { print "Game history position: ".$gameHistoryPosition."\n"; }

    # then set the position to the new top of the stack
    my @gameSnapshot = @{$gameHistory[$gameHistoryPosition]};

    if( $verbose) { print "Items found in snapshot: ".@gameSnapshot."\n"; }

    # restore the board position
    foreach my $position (@square) {
        $position->configure( -image => shift @gameSnapshot);
    }
    if( $verbose) { print "Board images loaded: ".@square."\n"; }
    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # restore the top kugel set
    foreach my $position (@topKugel) {
        $position->configure( -image => shift @gameSnapshot);
    }
    if( $verbose) { print "Top kugel images loaded: ".@topKugel."\n"; }
    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # set the number in the current set
    $#currentSet = shift @gameSnapshot;
    if( $verbose) { print "Current set #: ".@currentSet."\n"; }

    # restore the current set of kugels being used
    foreach my $position (@currentSet) {
        $position = shift @gameSnapshot;
    }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # restore the skip info, score, highest score, champion name
    $didSkip      = shift @gameSnapshot;
    $highestScore = shift @gameSnapshot;
    $champion     = shift @gameSnapshot;
    $score        = shift @gameSnapshot;

    if($verbose) { print "didSkip="     .$didSkip      ."\n"; }
    if($verbose) { print "highestScore=".$highestScore ."\n"; }
    if($verbose) { print "champion="    .$champion     ."\n"; }
    if($verbose) { print "score="       .$score        ."\n"; }

    if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

    # recalculate the number of squares filled
    $filled = 0;
    foreach my $position (@square) {
        unless( $position->cget(-image) == $currentSet[0] ) {
            $filled++;
        }
    }
    if( $verbose) { print "Filled squares: ".$filled."\n"; }

    # re-display the Filled button on the screen
    $numFilledLabel->configure(-text => (sprintf '%02d',
                                int($filled/($boardSizeX*$boardSizeY)*100))
                               ."\% Filled");

    # re-display score and highest score
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $highScoreLabel->configure(-text =>"$champion: ".(sprintf '%04d',$highestScore));
    $mainFrame->configure( -title => "Kugel Score: $score");
}

#
# save the entire game, it will save current position, as well as all the snapshots 
#
sub save_game_to_file {
    # game snapshots are Tk image objects (mostly), to save them on the 
    # disc, i will need to convert them to integer values that will be 
    # loaded into program again and converted to Tk image snapshots 
    # upom game reload from the same file

    if( $verbose) { print "save_game_to_file function\n"; }
    if( $verbose) { print "Number of snapshots in history: ".@gameHistory."\n"; }

    # array containing data to be stored
    my @data_to_store;

    foreach my $snapshot (@gameHistory) {
        my @gameSnapshot = @{$snapshot};

        # table containing board images from the snapshot
        # the number of squares is the same for ever snapshot, 
        # use dimansion of @square as argument to splice
        my @boardImages = splice @gameSnapshot, 0, @square;
        if( $verbose) { print "Board images: ".@boardImages."\n"; }

        # top kugel set
        my @topKugelImages = splice @gameSnapshot, 0, @topKugel;
        if( $verbose) { print "Top kugel images: ".@topKugelImages."\n"; }

        # number of images in the current set for this snapshot
        my $imagesInCurrentSet = (shift @gameSnapshot) + 1;
        if( $verbose) { print "Current set #: ".$imagesInCurrentSet."\n"; }

        # set of kugels being used for this snapshot
        my @currentSetImages = splice @gameSnapshot, 0, $imagesInCurrentSet;
        if( $verbose) { print "Current set images: ".@currentSetImages."\n"; }

        if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

        # skip info, score, highest score, champion name
        my $didSkipSnapshot      = shift @gameSnapshot;
        my $highestScoreSnapshot = shift @gameSnapshot;
        my $championSnapshot     = shift @gameSnapshot;
        my $scoreSnapshot        = shift @gameSnapshot;

        if($verbose) { print "didSkipSnapshot="     .$didSkipSnapshot      ."\n"; }
        if($verbose) { print "highestScoreSnapshot=".$highestScoreSnapshot ."\n"; }
        if($verbose) { print "championSnapshot="    .$championSnapshot     ."\n"; }
        if($verbose) { print "scoreSnapshot="       .$scoreSnapshot        ."\n"; }

        if( $verbose) { print "Items left in snapshot: ".@gameSnapshot."\n"; }

        # now that we have all the data from this snapshot, convert the
        # @boardImages and @topKugelImages arrays to arrays containing
        # integers, instead of Tk image objects
        my @boardImagesInteger = ();
        my @topKugelImagesInteger = ();

        # this variable should in the end show us that all images are recognized
        my $imagesRecognized = 0;
        for( my $i=0; $i<@boardImages; $i++ ) {
            for( my $j=0; $j<@currentSetImages; $j++ ) {
                if( $boardImages[$i]  == $currentSetImages[$j] ) {

                # save the image as integer
                push @boardImagesInteger, $j;

                    if($verbose) { print $i.":".$j." "; }
                    $imagesRecognized++;
                    next;
                }
            }
        }
      if($verbose) { print "\nRecognized ".$imagesRecognized." board images\n"; }
        unless( $imagesRecognized == @boardImages ) {
            &show_error("\nCoding error: \n\nUnrecognized images \nduring save operation.");
            return;
        }

        $imagesRecognized = 0;
        for( my $i=0; $i<@topKugelImages; $i++ ) {
            for( my $j=0; $j<@currentSetImages; $j++ ) {
                if( $topKugelImages[$i]  == $currentSetImages[$j] ) {
                # save the image as integer
                push @topKugelImagesInteger, $j;
                    if($verbose) { print $i.":".$j." "; }
                    $imagesRecognized++;
                    next;
              }
            }
        }
      if($verbose) { print "\nRecognized ".$imagesRecognized." top kugel images\n"; }
        unless( $imagesRecognized == @topKugelImages ) {
            &show_error("\nCoding error: \n\nUnrecognized top kugel images \nduring save operation.");
            return;
        }

      push @data_to_store, { 'board'              => \@boardImagesInteger,
                             'topKugel'           => \@topKugelImagesInteger,
                         'didSkip'            => $didSkipSnapshot,
                         'highestScore'       => $highestScoreSnapshot,
                         'champion'           => $championSnapshot,
                         'score'              => $scoreSnapshot,
                         'imagesInCurrentSet' => $imagesInCurrentSet }
    }

    # save the history position, so upon reload from the file
    # user can start from the very same undo/redo position
    if($verbose) { print "History position: ".$gameHistoryPosition."\n"; }

    my $fileName = localtime;
    $fileName =~ s/\s/_/g;

    unless( -f "$ENV{HOME}/.kugel" ) { 
        mkdir "$ENV{HOME}/.kugel", 0755; 
      if($verbose) { print "Created directory: $ENV{HOME}/.kugel\n"; }
    }
    my $fullPathFileName = $mainFrame->getSaveFile( -initialdir => "$ENV{HOME}/.kugel" ,
                                                    -initialfile => $fileName.".kgl",
                                                -defaultextension => '.kgl' );
    # if user canceled
    unless($fullPathFileName) { 
        if($verbose) { print "No file chosen. Nothing to do.\n"; }
      return;
    }

    if($verbose) { print "File name: ".$fullPathFileName."\n"; }

    unless( open(FILE, "> $fullPathFileName")) {
        &show_error("\nCould not open file\n$fullPathFileName");
      return;
    }

    # now we are ready to dump all this data to a file
    print FILE "Kugel game saved on ".localtime()."\n";
    print FILE "Board size X: " .$boardSizeX              ."\n";
    print FILE "Board size Y: " .$boardSizeY              ."\n";
    print FILE "Add by: "       .$addKugel                ."\n";
    print FILE "Delete by: "    .$delKugel                ."\n";
    print FILE "Using set of : ".$setOf                   ."\n";
    print FILE "Top kugels: "   .@topKugel                ."\n";

    foreach my $snapshot (@data_to_store) {

      print FILE "Snapshot ........................\n";
        print FILE "Board images:";
      foreach my $image (@{$snapshot->{'board'}}) {
          print FILE " ".$image;
      }
      print FILE "\n";

        print FILE "Top kugel images:";
      foreach my $image (@{$snapshot->{'topKugel'}}) {
          print FILE " ".$image;
      }
      print FILE "\n";

      print FILE "Did skip: ".$snapshot->{'didSkip'}."\n";
      print FILE "Highest score: ".$snapshot->{'highestScore'}."\n";
      print FILE "Champion: ".$snapshot->{'champion'}."\n";
      print FILE "Score: ".$snapshot->{'score'}."\n";
      print FILE "Images in current set: ".$snapshot->{'imagesInCurrentSet'}."\n";
    }

    close FILE;
}

sub load_game_from_file {
    unless( -f "$ENV{HOME}/.kugel" ) { 
        mkdir "$ENV{HOME}/.kugel", 0755; 
      if($verbose) { print "Created directory: $ENV{HOME}/.kugel\n"; }
    }
    my $fullPathFileName = $mainFrame->getOpenFile( -initialdir => "$ENV{HOME}/.kugel" ,
                                                -defaultextension => '.kgl' );
    unless($fullPathFileName) { 
        if($verbose) { print "No file chosen. Nothing to do.\n"; }
      return;
    }

    if($verbose) { print "File name: ".$fullPathFileName."\n"; }

    unless(open(FILE, $fullPathFileName)) {
        &show_error("\nCould not open file\n$fullPathFileName");
      return;
    }

    # array containing data to be stored
    my @data_to_restore;

    # this variable will keep track of the current snapshot being loaded
    my $snapshotNumber = 0;

    # this variable will be set if the dimensions (or # of kugels to add) has changed
    my $dimensionsChanged = 0;
    while(my $line = <FILE>) {
      # first few lines are global, not associated with particular snapshot
        if( $line =~ /Kugel game saved on / ) { 
          if($verbose) { print "Reading: ".$line; }
          next; 
        }
        if( $line =~ /Board size X: (.*)/ ) {
          # if the board size has changed
          if( $boardSizeX != $1 ) {
            $boardSizeX = $1;
                $dimensionsChanged = 1;
          }
            next;
      }
        if( $line =~ /Board size Y: (.*)/ ) {
          # if the board size has changed
          if( $boardSizeY != $1 ) {
            $boardSizeY = $1;
                $dimensionsChanged = 1;
          }
            next;
      }
        if( $line =~ /Add by: (.*)/ ) {
          # if the number of kugels to add has changed
          if( $addKugel != $1 ) {
            $addKugel = $1;
                $dimensionsChanged = 1;
          }
            next;
      }
        if( $line =~ /Delete by: (.*)/ ) {
          # if the number of kugels to delete has changed
          if( $delKugel != $1 ) {
            $delKugel = $1;
                $dimensionsChanged = 1;
          }
            next;
      }
        if( $line =~ /Using set of: (.*)/ ) {
          # if the number of kugels used in set has changed
          if( $setOf != $1 ) {
            $setOf = $1;
                $dimensionsChanged = 1;
          }
            next;
      }

        # read snapshot specific data
        if( $line =~ /Snapshot \.\.\.\.\.\./ ) { next; }
        if( $line =~ /Board images: (.*)/ ) { 
          my @board = split / /, $1;
          $data_to_restore[$snapshotNumber]->{'board'} = \@board;
          if($verbose) { print "Board elements loaded: ".@board."\n"; }
          next;
      }
      if( $line =~ /Top kugel images: (.*)/ ) {
          my @topKugel = split / /, $1;
          $data_to_restore[$snapshotNumber]->{'topKugel'} = \@topKugel;
          if($verbose) { print "Top kugel elements loaded: ".@topKugel."\n"; }
          next;
      }
        if( $line =~ /Did skip: (.*)/ ) { 
          $data_to_restore[$snapshotNumber]->{'didSkip'} = $1;
          if($verbose) { print "Did skip loaded: ".$data_to_restore[$snapshotNumber]->{'didSkip'}."\n"; }
          next;
      }
        if( $line =~ /Highest score: (.*)/ ) { 
          $data_to_restore[$snapshotNumber]->{'highestScore'} = $1;
          if($verbose) { print "Highest score loaded: ".$data_to_restore[$snapshotNumber]->{'highestScore'}."\n"; }
          next;
      }
        if( $line =~ /Champion: (.*)/ ) { 
          $data_to_restore[$snapshotNumber]->{'champion'} = $1;
          if($verbose) { print "Champion loaded: ".$data_to_restore[$snapshotNumber]->{'champion'}."\n"; }
          next;
      }
        if( $line =~ /Score: (.*)/ ) { 
          $data_to_restore[$snapshotNumber]->{'score'} = $1;
          if($verbose) { print "Score loaded: ".$data_to_restore[$snapshotNumber]->{'score'}."\n"; }
          next;
      }
        if( $line =~ /Images in current set: (.*)/ ) { 
          $data_to_restore[$snapshotNumber]->{'imagesInCurrentSet'} = $1;
          if($verbose) { print "Images in current set loaded: ".
                         $data_to_restore[$snapshotNumber]->{'imagesInCurrentSet'}."\n"; }

          # this line is the last in record, so skip to the next snapshot
          if($verbose) { print "Snapshot number ".$snapshotNumber." loaded\n"; }
          $snapshotNumber++;
          next;
      }
      if($verbose) { print "Unrecognized entry: ".$line; };
    }

    if($verbose) { print "Done reading file ".$fullPathFileName."\n"; }
    close FILE;

    # if the dimensions have changed, discard the previous history data
    if( $dimensionsChanged ) {
      # full restart will re-draw the board, and take the snapshot
      # the snapshot taken will be ignored after the history is deleted
      &full_restart;
      if($verbose) { print "New dimensions: boardSizeX=$boardSizeX, boardSizeY=$boardSizeY, ".
                           "addKugel=$addKugel, delKugel=$delKugel, setOf=$setOf\n"; }

        @gameHistory = ();
        $gameHistoryPosition = -1;
        if($verbose) { print "History deleted.\n"; }
    } else {
      # set the history position to the last element, before adding new entries
        $gameHistoryPosition = $#gameHistory; 
    }
    if($verbose) { print "Snapshots in history: ".@gameHistory."\n"; }

    # truncate the history array to the current position, after reload
    # user can not redo previously undoned moves
    $#gameHistory = $gameHistoryPosition;

    foreach my $snapshot (@data_to_restore) {
      # just an array representation of current snapshot
      # undo/redo is using array, not hash, wish I made
      # it a hash a while ago, but i did not remember that
      my @snapshotArray;
      foreach my $imageNumber (@{$snapshot->{'board'}}) {
          push @snapshotArray, $currentSet[$imageNumber]
      }
      foreach my $imageNumber (@{$snapshot->{'topKugel'}}) {
          push @snapshotArray, $currentSet[$imageNumber]
      }

      # do not change the current set of images used, leave it as is
      push @snapshotArray, $#currentSet ;
      push @snapshotArray, @currentSet;

      push @snapshotArray, $snapshot->{'didSkip'};
      push @snapshotArray, $snapshot->{'highestScore'};
      push @snapshotArray, $snapshot->{'champion'};
      push @snapshotArray, $snapshot->{'score'};

      # this item is ignored
      # push @snapshotArray, $snapshot->{'imagesInCurrentSet'};

      # add current snapshot 
      push @gameHistory, \@snapshotArray;

        # $#gameHistory will point to the last element in the @gameHistory array
        $gameHistoryPosition = $#gameHistory;

        if($verbose) { print "Added snapshot #: ".$gameHistoryPosition."\n"; }
    }

    # little trick, one step back in counting moves, plus the redo will re draw everything we need
    $gameHistoryPosition--;
    &redo_one_move;
}
